import re
import os
import codecs
import time
import stat
import mimetypes
import tempfile
import string
from collections import deque
from copy import deepcopy
from xml.sax.saxutils import escape as html_escape
from docxtpl import InlineImage, RichText
from docx.shared import Mm, Inches, Pt, Cm, Twips
import docx.opc.constants
from docx.oxml.section import CT_SectPr
from docx.oxml.table import CT_Tbl
import docx
from docxcompose.composer import Composer  # For fixing up images, etc when including docx files within templates
from docassemble.base.functions import server, this_thread, package_template_filename, get_config, roman
from docassemble.base.error import DAError
import docassemble.base.filter
import docassemble.base.pandoc
from docassemble.base.logger import logmessage
from bs4 import BeautifulSoup, NavigableString, Tag
from pikepdf import Pdf

zerowidth = '\u200B'

QPDF_PATH = 'qpdf'
NoneType = type(None)

DEFAULT_PAGE_WIDTH = '6.5in'

list_types = ['1', 'A', 'a', 'I', 'i']


def image_for_docx(fileref, question, tpl, width=None):
    if fileref.__class__.__name__ in ('DAFile', 'DAFileList', 'DAFileCollection', 'DALocalFile', 'DAStaticFile'):
        file_info = {'fullpath': fileref.path()}
    else:
        file_info = server.file_finder(fileref, question=question)
    if 'path' in file_info and 'extension' in file_info:
        docassemble.base.filter.convert_svg_to_png(file_info)
    if 'fullpath' not in file_info:
        return '[FILE NOT FOUND]'
    if width is not None:
        m = re.search(r'^([0-9\.]+) *([A-Za-z]*)', str(width))
        if m:
            amount = float(m.group(1))
            units = m.group(2).lower()
            if units in ['in', 'inches', 'inch']:
                the_width = Inches(amount)
            elif units in ['pt', 'pts', 'point', 'points']:
                the_width = Pt(amount)
            elif units in ['mm', 'millimeter', 'millimeters']:
                the_width = Mm(amount)
            elif units in ['cm', 'centimeter', 'centimeters']:
                the_width = Cm(amount)
            elif units in ['twp', 'twip', 'twips']:
                the_width = Twips(amount)
            else:
                the_width = Pt(amount)
        else:
            the_width = Inches(2)
    else:
        the_width = Inches(2)
    return InlineImage(tpl, file_info['fullpath'], the_width)


def transform_for_docx(text):
    if type(text) in (int, float, bool, NoneType):
        return text
    text = str(text)
    return text


def create_hyperlink(url, anchor_text, tpl):
    return InlineHyperlink(tpl, url, sanitize_xml(anchor_text))


class InlineHyperlink:

    def __init__(self, tpl, url, anchor_text):
        self.tpl = tpl
        self.url = url
        self.anchor_text = anchor_text

    def _insert_link(self):
        ref = self.tpl.docx._part.relate_to(self.url, docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True)
        return '</w:t></w:r><w:hyperlink r:id="%s"><w:r><w:rPr><w:rStyle w:val="InternetLink"/></w:rPr><w:t>%s</w:t></w:r></w:hyperlink><w:r><w:rPr></w:rPr><w:t xml:space="preserve">' % (ref, html_escape(self.anchor_text))

    def __str__(self):
        return self._insert_link()


def fix_subdoc(masterdoc, subdoc_info):
    """Fix the images, styles, references, shapes, etc of a subdoc"""
    subdoc = subdoc_info['subdoc']
    change_numbering = subdoc_info['change_numbering']
    composer = Composer(masterdoc)  # Using docxcompose
    composer.reset_reference_mapping()

    # This is the same as the docxcompose function, except it doesn't copy the elements over.
    # Copying the elements over is done by returning the subdoc XML in this function.
    # Both sd.subdocx and the master template file are changed with these functions.
    composer._create_style_id_mapping(subdoc)
    for element in subdoc.element.body:
        if isinstance(element, CT_SectPr):
            continue
        composer.add_referenced_parts(subdoc.part, masterdoc.part, element)
        composer.add_styles(subdoc, element)
        if change_numbering and not isinstance(element, CT_Tbl):
            try:
                composer.add_numberings(subdoc, element)
                composer.restart_first_numbering(subdoc, element)
            except:
                pass
        composer.add_images(subdoc, element)
        composer.add_shapes(subdoc, element)
        composer.add_footnotes(subdoc, element)
        composer.remove_header_and_footer_references(subdoc, element)

    composer.add_styles_from_other_parts(subdoc)
    composer.renumber_bookmarks()
    composer.renumber_docpr_ids()
    composer.fix_section_types(subdoc)


def include_docx_template(template_file, **kwargs):
    """Include the contents of one docx file inside another docx file."""
    use_jinja = kwargs.pop('_use_jinja2', True)
    if this_thread.evaluation_context is None:
        return 'ERROR: not in a docx file'
    if template_file.__class__.__name__ in ('DAFile', 'DAFileList', 'DAFileCollection', 'DALocalFile', 'DAStaticFile'):
        template_path = template_file.path()
    else:
        template_path = package_template_filename(template_file, package=this_thread.current_package)
    sd = this_thread.misc['docx_template'].new_subdoc()
    sd.subdocx = docx.Document(template_path)
    if not use_jinja:
        return sanitize_xml(str(sd))
    if '_inline' in kwargs:
        single_paragraph = True
        del kwargs['_inline']
    else:
        single_paragraph = False
    change_numbering = bool(kwargs.pop('change_numbering', True))

    # We need to keep a copy of the subdocs so we can fix up the master template in the end (in parse.py)
    # Given we're half way through processing the template, we can't fix the master template here
    # we have to do it in post
    if 'docx_subdocs' not in this_thread.misc:
        this_thread.misc['docx_subdocs'] = []
    this_thread.misc['docx_subdocs'].append({'subdoc': deepcopy(sd.subdocx), 'change_numbering': change_numbering})

    # Fix the subdocs before they are included in the template
    fix_subdoc(this_thread.misc['docx_template'], {'subdoc': sd.subdocx, 'change_numbering': change_numbering})

    first_paragraph = sd.subdocx.paragraphs[0]
    for key, val in kwargs.items():
        if hasattr(val, 'instanceName'):
            the_repr = val.instanceName
        else:
            the_repr = '_codecs.decode(_array.array("b", "' + re.sub(r'\n', '', codecs.encode(bytearray(val, encoding='utf-8'), 'base64').decode()) + '".encode()), "base64").decode()'
        first_paragraph.insert_paragraph_before(str("{%%p set %s = %s %%}" % (key, the_repr)))
    if 'docx_include_count' not in this_thread.misc:
        this_thread.misc['docx_include_count'] = 0
    this_thread.misc['docx_include_count'] += 1
    if single_paragraph:
        return re.sub(r'<w:p[^>]*>\s*(.*)</w:p>\s*', r'\1', str(first_paragraph._p.xml), flags=re.DOTALL)
    return sd


def get_children(descendants, parsed):
    subelement = False
    descendants_buff = deque()
    if descendants is None:
        return descendants_buff
    if isinstance(descendants, NavigableString):
        parsed.append(descendants)
    else:
        for child in descendants.children:
            if child.name is None:
                if subelement is False:
                    parsed.append(child)
                else:
                    descendants_buff.append(child)
            else:
                if subelement is False:
                    subelement = True
                    descendants_buff.append(child)
                else:
                    descendants_buff.append(child)
    descendants_buff.reverse()
    return descendants_buff


def html_linear_parse(soup):
    html_tag = soup.html
    descendants = deque()
    descendants.appendleft(html_tag)
    parsed = deque()
    while len(list(descendants)) > 0:
        child = descendants.popleft()
        from_children = get_children(child, parsed)
        descendants.extendleft(from_children)
    return parsed


def Alpha(number):
    multiplier = int((number - 1) / 26)
    indexno = (number - 1) % 26
    return string.ascii_uppercase[indexno] * (multiplier + 1)


def alpha(number):
    multiplier = int((number - 1) / 26)
    indexno = (number - 1) % 26
    return string.ascii_lowercase[indexno] * (multiplier + 1)


def Roman_Numeral(number):
    return roman((number - 1) % 4000, case='upper')


def roman_numeral(number):
    return roman((number - 1) % 4000, case='lower')


class SoupParser:

    def __init__(self, tpl):
        self.paragraphs = [{'params': {'style': 'p', 'indentation': 0, 'list_number': 1}, 'runs': [RichText('')]}]
        self.current_paragraph = self.paragraphs[-1]
        self.run = self.current_paragraph['runs'][-1]
        self.bold = False
        self.center = False
        self.list_number = 1
        self.list_type = list_types[-1]
        self.italic = False
        self.underline = False
        self.strike = False
        self.indentation = 0
        self.style = 'p'
        self.still_new = True
        self.size = None
        self.charstyle = None
        self.color = None
        self.tpl = tpl

    def new_paragraph(self, classes, styles):
        if self.still_new:
            # logmessage("new_paragraph is still new and style is " + self.style + " and indentation is " + str(self.indentation))
            self.current_paragraph['params']['style'] = self.style
            self.current_paragraph['params']['indentation'] = self.indentation
            self.set_attribs(classes, styles)
            self.list_number += 1
            return
        # logmessage("new_paragraph where style is " + self.style + " and indentation is " + str(self.indentation))
        self.current_paragraph = {'params': {'style': self.style, 'indentation': self.indentation, 'list_number': self.list_number}, 'runs': [RichText('')]}
        self.set_attribs(classes, styles)
        self.list_number += 1
        self.paragraphs.append(self.current_paragraph)
        self.run = self.current_paragraph['runs'][-1]
        self.still_new = True

    def set_attribs(self, classes, styles):
        if 'dacenter' in classes:
            self.current_paragraph['params']['align'] = 'center'
        elif 'daflushright' in classes:
            self.current_paragraph['params']['align'] = 'end'
        else:
            self.current_paragraph['params']['align'] = 'start'
        if len(classes):
            if 'daspacingtight' in classes:
                self.current_paragraph['params']['spacing'] = 240
                self.current_paragraph['params']['after'] = 0
            elif 'daspacingsingle' in classes:
                self.current_paragraph['params']['spacing'] = 240
                self.current_paragraph['params']['after'] = 240
            elif 'daspacingdouble' in classes:
                self.current_paragraph['params']['spacing'] = 480
                self.current_paragraph['params']['after'] = 0
            elif 'daspacingoneandahalf' in classes:
                self.current_paragraph['params']['spacing'] = 260
                self.current_paragraph['params']['after'] = 0
            elif 'daspacingtriple' in classes:
                self.current_paragraph['params']['spacing'] = 700
                self.current_paragraph['params']['after'] = 0
        if styles:
            m = re.search(r'margin-left:([0-9\.]+)px', styles)
            if m:
                self.current_paragraph['params']['leftindent'] = 20 * int(m.group(1))
            m = re.search(r'margin-right:([0-9\.]+)px', styles)
            if m:
                self.current_paragraph['params']['rightindent'] = 20 * int(m.group(1))
            m = re.search(r'text-indent:([0-9\.]+)px', styles)
            if m:
                self.current_paragraph['params']['firstline'] = 20 * int(m.group(1))

    def __str__(self):
        output = ''
        for para in self.paragraphs:
            # logmessage("Got a paragraph where style is " + para['params']['style'] + " and indentation is " + str(para['params']['indentation']))
            output += '<w:p><w:pPr><w:pStyle w:val="Normal"/>'
            if para['params']['align'] == 'center':
                output += '<w:jc w:val="center"/>'
            elif para['params']['align'] == 'end':
                output += '<w:jc w:val="end"/>'
            if 'spacing' in para['params']:
                output += '<w:spacing w:before="0" w:after="' + str(para['params']['after']) + '" w:line="' + str(para['params']['spacing']) + '" w:lineRule="auto" w:beforeAutospacing="0" w:afterAutospacing="0"/>'
            if para['params']['style'] == 'ul' or para['params']['style'].startswith('ol'):
                if 'leftindent' in para['params']:
                    left_indent = para['params']['leftindent']
                else:
                    left_indent = 36*para['params']['indentation']
                if 'rightindent' in para['params']:
                    right_indent = para['params']['rightindent']
                else:
                    right_indent = 0
                output += '<w:ind w:left="' + str(left_indent) + '" w:right="' + str(right_indent) + '" w:hanging="0"/>'
            elif para['params']['style'] == 'blockquote':
                if 'spacing' not in para['params']:
                    output += '<w:spacing w:before="0" w:after="240" w:line="240" w:lineRule="auto" w:beforeAutospacing="0" w:afterAutospacing="0"/>'
                output += '<w:ind w:left="' + str(36*para['params']['indentation']) + '" w:right="' + str(36*para['params']['indentation']) + '" w:hanging="0"/>'
            elif 'leftindent' in para['params'] or 'rightindent' in para['params'] or 'firstline' in para['params']:
                if 'leftindent' in para['params']:
                    left_indent = para['params']['leftindent']
                else:
                    left_indent = 0
                if 'rightindent' in para['params']:
                    right_indent = para['params']['rightindent']
                else:
                    right_indent = 0
                if 'firstline' in para['params']:
                    first_line = para['params']['firstline']
                else:
                    first_line = 0
                output += '<w:ind w:left="' + str(left_indent) + '" w:right="' + str(right_indent) + '" w:hanging="0" w:firstLine="' + str(first_line) + '"/>'
            output += '<w:rPr></w:rPr></w:pPr>'
            if para['params']['style'] == 'ul':
                output += str(RichText("•\t"))
            if para['params']['style'] == 'ol1':
                output += str(RichText(str(para['params']['list_number']) + ".\t"))
            elif para['params']['style'] == 'olA':
                output += str(RichText(Alpha(para['params']['list_number']) + ".\t"))
            elif para['params']['style'] == 'ola':
                output += str(RichText(alpha(para['params']['list_number']) + ".\t"))
            elif para['params']['style'] == 'olI':
                output += str(RichText(Roman_Numeral(para['params']['list_number']) + ".\t"))
            elif para['params']['style'] == 'oli':
                output += str(RichText(roman_numeral(para['params']['list_number']) + ".\t"))
            for run in para['runs']:
                output += str(run)
            output += '</w:p>'
        return output

    def start_link(self, url):
        ref = self.tpl.docx._part.relate_to(url, docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True)
        self.current_paragraph['runs'].append('<w:hyperlink r:id="%s">' % (ref, ))
        self.new_run()
        self.still_new = False

    def end_link(self):
        self.current_paragraph['runs'].append('</w:hyperlink>')
        self.new_run()
        self.still_new = False

    def new_run(self):
        self.current_paragraph['runs'].append(RichText(''))
        self.run = self.current_paragraph['runs'][-1]

    def traverse(self, elem):
        for part in elem.contents:
            if isinstance(part, NavigableString):
                self.run.add(str(part), italic=self.italic, bold=self.bold, underline=self.underline, strike=self.strike, size=self.size, style=self.charstyle, color=self.color)
                self.still_new = False
            elif isinstance(part, Tag):
                # logmessage("Part name is " + str(part.name))
                if part.name == 'p':
                    if 'class' in part.attrs:
                        classes = part.attrs['class']
                    else:
                        classes = []
                    if 'style' in part.attrs:
                        styles = part.attrs['style']
                    else:
                        styles = ""
                    self.new_paragraph(classes, styles)
                    if 'dabold' in classes:
                        self.bold = True
                    self.traverse(part)
                    if 'dabold' in classes:
                        self.bold = False
                elif part.name == 'li':
                    if 'class' in part.attrs:
                        classes = part.attrs['class']
                    else:
                        classes = []
                    if 'style' in part.attrs:
                        styles = part.attrs['style']
                    else:
                        styles = ""
                    self.new_paragraph(classes, styles)
                    self.traverse(part)
                elif part.name == 'ul':
                    # logmessage("Entering a UL")
                    oldstyle = self.style
                    self.style = 'ul'
                    self.indentation += 10
                    self.traverse(part)
                    self.indentation -= 10
                    self.style = oldstyle
                    # logmessage("Leaving a UL")
                elif part.name == 'ol':
                    # logmessage("Entering a OL")
                    oldstyle = self.style
                    oldlistnumber = self.list_number
                    oldlisttype = self.list_type
                    if part.get('type', None) in list_types:
                        self.list_type = part['type']
                    else:
                        self.list_type = list_types[(list_types.index(self.list_type) + 1) % 5]
                    try:
                        self.list_number = int(part.get('start', 1))
                    except:
                        self.list_number = 1
                    self.style = 'ol' + self.list_type
                    self.indentation += 10
                    self.traverse(part)
                    self.indentation -= 10
                    self.list_type = oldlisttype
                    self.list_number = oldlistnumber
                    self.style = oldstyle
                    # logmessage("Leaving a OL")
                elif part.name == 'strong':
                    self.bold = True
                    self.traverse(part)
                    self.bold = False
                elif part.name == 'em':
                    self.italic = True
                    self.traverse(part)
                    self.italic = False
                elif part.name == 'strike':
                    self.strike = True
                    self.traverse(part)
                    self.strike = False
                elif part.name == 'u':
                    self.underline = True
                    self.traverse(part)
                    self.underline = False
                elif part.name == 'blockquote':
                    oldstyle = self.style
                    self.style = 'blockquote'
                    self.indentation += 20
                    self.traverse(part)
                    self.indentation -= 20
                    self.style = oldstyle
                elif re.match(r'h[1-6]', part.name):
                    oldsize = self.size
                    self.size = 60 - ((int(part.name[1]) - 1) * 10)
                    if 'class' in part.attrs:
                        classes = part.attrs['class']
                    else:
                        classes = []
                    if 'style' in part.attrs:
                        styles = part.attrs['style']
                    else:
                        styles = ""
                    self.new_paragraph(classes, styles)
                    self.bold = True
                    self.traverse(part)
                    self.bold = False
                    self.size = oldsize
                elif part.name == 'a':
                    self.start_link(part['href'])
                    if self.tpl.da_hyperlink_style:
                        self.charstyle = self.tpl.da_hyperlink_style
                    else:
                        self.underline = True
                        self.color = '#0000ff'
                    self.traverse(part)
                    if self.tpl.da_hyperlink_style:
                        self.charstyle = None
                    else:
                        self.underline = False
                        self.color = None
                    self.end_link()
                elif part.name == 'br':
                    self.run.add("\n", italic=self.italic, bold=self.bold, underline=self.underline, strike=self.strike, size=self.size, style=self.charstyle, color=self.color)
                    self.still_new = False
            else:
                logmessage("Encountered a " + part.__class__.__name__)


class InlineSoupParser:

    def __init__(self, tpl):
        self.runs = [RichText('')]
        self.run = self.runs[-1]
        self.bold = False
        self.italic = False
        self.underline = False
        self.indentation = 0
        self.style = 'p'
        self.strike = False
        self.size = None
        self.charstyle = None
        self.color = None
        self.tpl = tpl
        self.at_start = True
        self.list_number = 1
        self.list_type = list_types[-1]

    def new_paragraph(self):
        if self.at_start:
            self.at_start = False
        else:
            self.run.add("\n", italic=self.italic, bold=self.bold, underline=self.underline, strike=self.strike, size=self.size, style=self.charstyle, color=self.color)
        if self.indentation:
            self.run.add("\t" * self.indentation)
        if self.style == 'ul':
            self.run.add("•\t")
        if self.style == 'ol1':
            self.run.add(str(self.list_number) + ".\t")
            self.list_number += 1
        elif self.style == 'olA':
            self.run.add(Alpha(self.list_number) + ".\t")
            self.list_number += 1
        elif self.style == 'ola':
            self.run.add(alpha(self.list_number) + ".\t")
            self.list_number += 1
        elif self.style == 'olI':
            self.run.add(Roman_Numeral(self.list_number) + ".\t")
            self.list_number += 1
        elif self.style == 'oli':
            self.run.add(roman_numeral(self.list_number) + ".\t")
            self.list_number += 1
        # else:
        #     self.list_number = 1

    def __str__(self):
        output = ''
        for run in self.runs:
            output += str(run)
        return output

    def start_link(self, url):
        ref = self.tpl.docx._part.relate_to(url, docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True)
        self.runs.append('<w:hyperlink r:id="%s">' % (ref, ))
        self.new_run()

    def end_link(self):
        self.runs.append('</w:hyperlink>')
        self.new_run()

    def new_run(self):
        self.runs.append(RichText(''))
        self.run = self.runs[-1]

    def traverse(self, elem):
        for part in elem.contents:
            if isinstance(part, NavigableString):
                self.run.add(str(part), italic=self.italic, bold=self.bold, underline=self.underline, strike=self.strike, size=self.size, style=self.charstyle, color=self.color)
            elif isinstance(part, Tag):
                if part.name in ('p', 'blockquote'):
                    self.new_paragraph()
                    self.traverse(part)
                elif part.name == 'li':
                    self.new_paragraph()
                    self.traverse(part)
                elif part.name == 'ul':
                    oldstyle = self.style
                    self.style = 'ul'
                    self.indentation += 1
                    self.traverse(part)
                    self.indentation -= 1
                    self.style = oldstyle
                elif part.name == 'ol':
                    oldstyle = self.style
                    oldlistnumber = self.list_number
                    oldlisttype = self.list_type
                    if part.get('type', None) in list_types:
                        self.list_type = part['type']
                    else:
                        self.list_type = list_types[(list_types.index(self.list_type) + 1) % 5]
                    try:
                        self.list_number = int(part.get('start', 1))
                    except:
                        self.list_number = 1
                    self.style = 'ol' + self.list_type
                    self.indentation += 1
                    self.traverse(part)
                    self.indentation -= 1
                    self.list_type = oldlisttype
                    self.list_number = oldlistnumber
                    self.style = oldstyle
                elif part.name == 'strong':
                    self.bold = True
                    self.traverse(part)
                    self.bold = False
                elif part.name == 'em':
                    self.italic = True
                    self.traverse(part)
                    self.italic = False
                elif part.name == 'strike':
                    self.strike = True
                    self.traverse(part)
                    self.strike = False
                elif part.name == 'u':
                    self.underline = True
                    self.traverse(part)
                    self.underline = False
                elif re.match(r'h[1-6]', part.name):
                    oldsize = self.size
                    self.size = 60 - ((int(part.name[1]) - 1) * 10)
                    self.bold = True
                    self.traverse(part)
                    self.bold = False
                    self.size = oldsize
                elif part.name == 'a':
                    self.start_link(part['href'])
                    if self.tpl.da_hyperlink_style:
                        self.charstyle = self.tpl.da_hyperlink_style
                    else:
                        self.underline = True
                        self.color = '#0000ff'
                    self.traverse(part)
                    if self.tpl.da_hyperlink_style:
                        self.charstyle = None
                    else:
                        self.underline = False
                        self.color = None
                    self.end_link()
                elif part.name == 'br':
                    self.run.add("\n", italic=self.italic, bold=self.bold, underline=self.underline, strike=self.strike, size=self.size, style=self.charstyle, color=self.color)
            else:
                logmessage("Encountered a " + part.__class__.__name__)


def inline_markdown_to_docx(text, question, tpl):
    old_context = docassemble.base.functions.this_thread.evaluation_context
    docassemble.base.functions.this_thread.evaluation_context = None
    try:
        text = str(text)
    except:
        docassemble.base.functions.this_thread.evaluation_context = old_context
        raise
    docassemble.base.functions.this_thread.evaluation_context = old_context
    source_code = docassemble.base.filter.markdown_to_html(text, do_terms=False)
    source_code = re.sub(r"\n", ' ', source_code)
    source_code = re.sub(r">\s+<", '><', source_code)
    soup = BeautifulSoup('<html>' + source_code + '</html>', 'html.parser')
    parser = InlineSoupParser(tpl)
    for elem in soup.find_all(recursive=False):
        parser.traverse(elem)
    output = str(parser)
    return docassemble.base.filter.docx_template_filter(output, question=question, replace_newlines=False)


def markdown_to_docx(text, question, tpl):
    old_context = docassemble.base.functions.this_thread.evaluation_context
    docassemble.base.functions.this_thread.evaluation_context = None
    try:
        text = str(text)
    except:
        docassemble.base.functions.this_thread.evaluation_context = old_context
        raise
    docassemble.base.functions.this_thread.evaluation_context = old_context
    if get_config('new markdown to docx', False):
        source_code = docassemble.base.filter.markdown_to_html(text, do_terms=False)
        source_code = re.sub(r"\n", ' ', source_code)
        source_code = re.sub(r">\s+<", '><', source_code)
        soup = BeautifulSoup('<html>' + source_code + '</html>', 'html.parser')
        parser = SoupParser(tpl)
        for elem in soup.find_all(recursive=False):
            parser.traverse(elem)
        output = str(parser)
        # logmessage(output)
        return docassemble.base.filter.docx_template_filter(output, question=question)
    return inline_markdown_to_docx(text, question, tpl)


def pdf_pages(file_info, width):
    output = ''
    if width is None:
        width = DEFAULT_PAGE_WIDTH
    if not os.path.isfile(file_info['path'] + '.pdf'):
        if file_info['extension'] in ('rtf', 'doc', 'odt') and not os.path.isfile(file_info['path'] + '.pdf'):
            server.fg_make_pdf_for_word_path(file_info['path'], file_info['extension'])
    if 'pages' not in file_info:
        try:
            with Pdf.open(file_info['path'] + '.pdf') as reader:
                file_info['pages'] = len(reader.pages)
        except:
            file_info['pages'] = 1
    max_pages = 1 + int(file_info['pages'])
    formatter = '%0' + str(len(str(max_pages))) + 'd'
    for page in range(1, max_pages):
        page_file = {}
        test_path = file_info['path'] + 'page-in-progress'
        if os.path.isfile(test_path):
            while (os.path.isfile(test_path) and time.time() - os.stat(test_path)[stat.ST_MTIME]) < 30:
                if not os.path.isfile(test_path):
                    break
                time.sleep(1)
        page_file['extension'] = 'png'
        page_file['path'] = file_info['path'] + 'page-' + formatter % page
        page_file['fullpath'] = page_file['path'] + '.png'
        if not os.path.isfile(page_file['fullpath']):
            server.fg_make_png_for_pdf_path(file_info['path'] + '.pdf', 'page')
        if os.path.isfile(page_file['fullpath']):
            output += str(image_for_docx(docassemble.base.functions.DALocalFile(page_file['fullpath']), docassemble.base.functions.this_thread.current_question, docassemble.base.functions.this_thread.misc.get('docx_template', None), width=width))
        else:
            output += "[Error including page image]"
        output += ' '
    return output


def concatenate_files(path_list):
    new_path_list = []
    for path in path_list:
        mimetype, encoding = mimetypes.guess_type(path)  # pylint: disable=unused-variable
        if mimetype in ('application/rtf', 'application/msword', 'application/vnd.oasis.opendocument.text'):
            new_docx_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".docx", delete=False)
            if mimetype == 'application/rtf':
                ext = 'rtf'
            elif mimetype == 'application/msword':
                ext = 'doc'
            elif mimetype == 'application/vnd.oasis.opendocument.text':
                ext = 'odt'
            docassemble.base.pandoc.convert_file(path, new_docx_file.name, ext, 'docx')
            new_path_list.append(new_docx_file.name)
        elif mimetype == 'application/vnd.openxmlformats-officedocument.wordprocessingml.document':
            new_path_list.append(path)
    if len(new_path_list) == 0:
        raise DAError("concatenate_files: no valid files to concatenate")
    if len(new_path_list) == 1:
        return new_path_list[0]
    composer = Composer(docx.Document(new_path_list[0]))
    for indexno in range(1, len(new_path_list)):
        composer.append(docx.Document(new_path_list[indexno]))
    docx_file = tempfile.NamedTemporaryFile(prefix="datemp", mode="wb", suffix=".docx", delete=False)
    composer.save(docx_file.name)
    return docx_file.name


def sanitize_xml(text):
    return re.sub(r'{([{%#])', '{' + zerowidth + r'\1', re.sub(r'([}%#])}', r'\1' + zerowidth + '}', text))



import os
import mimetypes
import datetime
from botocore.errorfactory import ClientError
import boto3

epoch = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=datetime.timezone.utc)


class s3object:

    def __init__(self, s3_config):
        self.upload_args = {}
        self.download_args = {}
        if 'server side encryption' in s3_config and isinstance(s3_config['server side encryption'], dict):
            if s3_config['server side encryption'].get('algorithm'):
                self.upload_args['ServerSideEncryption'] = s3_config['server side encryption']['algorithm']
            if s3_config['server side encryption'].get('customer algorithm'):
                self.upload_args['SSECustomerAlgorithm'] = s3_config['server side encryption']['customer algorithm']
                self.download_args['SSECustomerAlgorithm'] = s3_config['server side encryption']['customer algorithm']
            if s3_config['server side encryption'].get('customer key'):
                self.upload_args['SSECustomerKey'] = s3_config['server side encryption']['customer key']
                self.download_args['SSECustomerKey'] = s3_config['server side encryption']['customer key']
            if s3_config['server side encryption'].get('KMS key ID'):
                self.upload_args['SSEKMSKeyId'] = s3_config['server side encryption']['KMS key ID']
        if 'access key id' in s3_config and s3_config['access key id'] is not None:
            self.conn = boto3.resource('s3', region_name=s3_config.get('region', None),
                                       aws_access_key_id=s3_config['access key id'],
                                       aws_secret_access_key=s3_config['secret access key'],
                                       endpoint_url=s3_config.get('endpoint url', None))
            self.client = boto3.client('s3', region_name=s3_config.get('region', None),
                                       aws_access_key_id=s3_config['access key id'],
                                       aws_secret_access_key=s3_config['secret access key'],
                                       endpoint_url=s3_config.get('endpoint url', None))
        else:
            self.conn = boto3.resource('s3', region_name=s3_config.get('region', None),
                                       endpoint_url=s3_config.get('endpoint url', None))
            self.client = boto3.client('s3', region_name=s3_config.get('region', None),
                                       endpoint_url=s3_config.get('endpoint url', None))
        self.bucket = self.conn.Bucket(s3_config['bucket'])
        self.bucket_name = s3_config['bucket']

    def get_key(self, key_name):
        return s3key(self, self.conn.Object(self.bucket_name, key_name))

    def search_key(self, key_name):
        for key in self.bucket.objects.filter(Prefix=key_name, Delimiter='/'):
            if key.key == key_name:
                return s3key(self, self.conn.Object(self.bucket_name, key.key))
        return None

    def list_keys(self, prefix):
        output = []
        for obj in self.bucket.objects.filter(Prefix=prefix):
            new_key = s3key(self, obj)
            output.append(new_key)
        return output


class s3key:

    def __init__(self, s3_object, key_obj):
        self.s3_object = s3_object
        self.key_obj = key_obj
        self.name = key_obj.key
        if self.key_obj.__class__.__name__ == 's3.ObjectSummary':
            self.size = self.key_obj.size
            self.last_modified = self.key_obj.last_modified
            self.does_exist = True
        elif self.exists():
            self.does_exist = True
        else:
            self.does_exist = False

    def get_contents_as_string(self):
        resp = self.key_obj.get(**self.s3_object.download_args)
        self.size = resp['ContentLength']
        self.content_type = resp['ContentType']
        self.last_modified = resp['LastModified']
        return resp['Body'].read().decode()

    def exists(self):
        try:
            resp = self.s3_object.client.head_object(Bucket=self.s3_object.bucket_name, Key=self.name, **self.s3_object.download_args)
            self.size = resp['ContentLength']
            self.content_type = resp['ContentType']
            self.last_modified = resp['LastModified']
        except ClientError:
            return False
        return True

    def delete(self):
        self.key_obj.delete()
        self.does_exist = False

    def get_epoch_modtime(self):
        return (self.last_modified - epoch).total_seconds()

    def get_contents_to_filename(self, filename):
        self.s3_object.conn.Bucket(self.s3_object.bucket_name).download_file(self.name, filename, ExtraArgs=self.s3_object.download_args)
        secs = (self.last_modified - epoch).total_seconds()
        os.utime(filename, (secs, secs))

    def set_contents_from_filename(self, filename):
        if hasattr(self, 'content_type') and self.content_type is not None:
            mimetype = self.content_type
        else:
            mimetype, encoding = mimetypes.guess_type(filename)  # pylint: disable=unused-variable
        if self.key_obj.__class__.__name__ == 's3.ObjectSummary':
            self.key_obj = self.s3_object.conn.Object(self.s3_object.bucket_name, self.name)
        if mimetype is not None:
            self.key_obj.upload_file(filename, ExtraArgs={'ContentType': mimetype, **self.s3_object.upload_args})
        else:
            self.key_obj.upload_file(filename, ExtraArgs=self.s3_object.upload_args)
        resp = self.s3_object.client.head_object(Bucket=self.s3_object.bucket_name, Key=self.name, **self.s3_object.download_args)
        self.size = resp['ContentLength']
        self.content_type = resp['ContentType']
        self.last_modified = resp['LastModified']
        secs = (self.last_modified - epoch).total_seconds()
        os.utime(filename, (secs, secs))

    def set_contents_from_string(self, text):
        if hasattr(self, 'content_type') and self.content_type is not None:
            self.key_obj.put(Body=bytes(text, encoding='utf-8'), ContentType=self.content_type, **self.s3_object.upload_args)
        else:
            self.key_obj.put(Body=bytes(text, encoding='utf-8'), **self.s3_object.upload_args)
        resp = self.s3_object.client.head_object(Bucket=self.s3_object.bucket_name, Key=self.name, **self.s3_object.download_args)
        self.size = resp['ContentLength']
        self.content_type = resp['ContentType']
        self.last_modified = resp['LastModified']

    def generate_url(self, expires, content_type=None, display_filename=None, inline=False):
        params = {'Bucket': self.s3_object.bucket_name, 'Key': self.key_obj.key}
        if content_type is not None:
            params['ResponseContentType'] = content_type
        if display_filename is not None:
            params['ResponseContentDisposition'] = "attachment; filename=" + display_filename
        if inline:
            params['ResponseContentDisposition'] = "inline"
        return self.s3_object.client.generate_presigned_url(
            ClientMethod='get_object',
            Params=params,
            ExpiresIn=expires)



# XXX these were copied as a quick fix; need to address this soon
class TAB:
    pass

# XXX these were copied as a quick fix; need to address this soon
class LINE:
    pass

class RawCode:
    def __init__(self, data):
        self.Data = data



import re


valid_variable_match = re.compile(r'^[^\d][A-Za-z0-9\_]*$')
match_brackets_or_dot = re.compile(r'(\[.+?\]|\.[a-zA-Z_][a-zA-Z0-9_]*)')

class DAIndexError(IndexError):
    pass


class DAAttributeError(AttributeError):
    pass


class DAException(Exception):
    pass


class DAError(Exception):

    def __init__(self, value, code=501):
        self.value = value
        self.error_code = code
        super().__init__(value)

    def __str__(self):
        return str(self.value)


class DANotFoundError(Exception):
    pass


class DAValidationError(Exception):
    """This is an Exception object that is used when raising an exception inside input validation code."""

    def __init__(self, *pargs, field=None):
        self.field = field
        super().__init__(*pargs)


class CodeExecute(Exception):

    def __init__(self, compute, question):
        if isinstance(compute, list):
            self.compute = "\n".join(compute)
        else:
            self.compute = compute
        self.question = question
        super().__init__()


class ForcedReRun(Exception):
    pass


class LazyNameError(NameError):
    pass


class DANameError(NameError):
    pass


def invalid_variable_name(varname):
    if not isinstance(varname, str):
        return True
    if re.search(r'[\n\r\(\)\{\}\*\^\#]', varname):
        return True
    varname = re.sub(r'[\.\[].*', '', varname)
    if not valid_variable_match.match(varname):
        return True
    return False


def intrinsic_name_of(var_name, the_user_dict):
    from docassemble.base.util import DAObject  # pylint: disable=import-outside-toplevel
    expression_as_list = [x for x in match_brackets_or_dot.split(var_name) if x != '']
    n = len(expression_as_list)
    i = n
    while i > 0:
        try:
            item = eval(var_name, the_user_dict)
            if isinstance(item, DAObject) and item.has_nonrandom_instance_name:
                var_name = item.instanceName
                break
        except:
            pass
        i -= 1
        var_name = ''.join(expression_as_list[0:i])
    return var_name + (''.join(expression_as_list[i:n]))


class ForcedNameError(NameError):

    def __init__(self, *pargs, **kwargs):
        super().__init__()
        the_args = list(pargs)
        if len(the_args) == 0:
            raise DAError("ForcedNameError must have at least one argument")
        the_context = {}
        the_user_dict = kwargs.get('user_dict', {})
        for var_name in ('x', 'i', 'j', 'k', 'l', 'm', 'n'):
            if var_name in the_user_dict:
                the_context[var_name] = the_user_dict[var_name]
        first_is_plain = bool(isinstance(the_args[0], str))
        self.next_action = []
        evaluate = kwargs.get('evaluate', False)
        while len(the_args) > 0:
            arg = the_args.pop(0)
            if isinstance(arg, dict):
                if (len(arg.keys()) == 2 and 'action' in arg and 'arguments' in arg) or (len(arg.keys()) == 1 and 'action' in arg):
                    arg['context'] = the_context
                    self.set_action(arg)
                elif len(arg) == 1 and ('undefine' in arg or 'invalidate' in arg or 'recompute' in arg or 'set' in arg or 'follow up' in arg):
                    if 'set' in arg:
                        if isinstance(arg['set'], dict):
                            arg['set'] = [arg['set']]
                        if not isinstance(arg['set'], list):
                            raise DAError("force_ask: the set statement must refer to a list.")
                        clean_list = []
                        for the_dict in arg['set']:
                            if not isinstance(the_dict, dict):
                                raise DAError("force_ask: a set command must refer to a list of dicts.")
                            for the_var, the_val in the_dict.items():
                                if not isinstance(the_var, str):
                                    raise DAError("force_ask: a set command must refer to a list of dicts with keys as variable names.  ")
                                the_var_stripped = the_var.strip()
                                if invalid_variable_name(the_var_stripped):
                                    raise DAError("force_ask: missing or invalid variable name " + repr(the_var) + ".")
                                clean_list.append([the_var_stripped, the_val])
                        self.set_action({'action': '_da_set', 'arguments': {'variables': clean_list}, 'context': the_context})
                    if 'follow up' in arg:
                        if isinstance(arg['follow up'], str):
                            arg['follow up'] = [arg['follow up']]
                        if not isinstance(arg['follow up'], list):
                            raise DAError("force_ask: the follow up statement must refer to a list.")
                        for var in arg['follow up']:
                            if not isinstance(var, str):
                                raise DAError("force_ask: invalid variable name " + repr(var) + " in follow up.")
                            var_saveas = var.strip()
                            if invalid_variable_name(var_saveas):
                                raise DAError("force_ask: missing or invalid variable name " + repr(var_saveas) + ".")
                            if evaluate:
                                var = intrinsic_name_of(var, the_user_dict)
                            self.set_action({'action': var, 'arguments': {}, 'context': the_context})
                    for command in ('undefine', 'invalidate', 'recompute'):
                        if command not in arg:
                            continue
                        if isinstance(arg[command], str):
                            arg[command] = [arg[command]]
                        if not isinstance(arg[command], list):
                            raise DAError("force_ask: the " + command + " statement must refer to a list.  ")
                        clean_list = []
                        for undef_var in arg[command]:
                            if not isinstance(undef_var, str):
                                raise DAError("force_ask: invalid variable name " + repr(undef_var) + " in " + command + ".")
                            undef_saveas = undef_var.strip()
                            if invalid_variable_name(undef_saveas):
                                raise DAError("force_ask: missing or invalid variable name " + repr(undef_saveas) + ".")
                            if evaluate:
                                undef_saveas = intrinsic_name_of(undef_saveas, the_user_dict)
                            clean_list.append(undef_saveas)
                        if command == 'invalidate':
                            self.set_action({'action': '_da_invalidate', 'arguments': {'variables': clean_list}, 'context': the_context})
                        else:
                            self.set_action({'action': '_da_undefine', 'arguments': {'variables': clean_list}, 'context': the_context})
                        if command == 'recompute':
                            self.set_action({'action': '_da_compute', 'arguments': {'variables': clean_list}, 'context': the_context})
                else:
                    raise DAError("Dictionaries passed to force_ask must have keys of 'action' and 'argument' only.")
            else:
                if evaluate:
                    arg = intrinsic_name_of(arg, the_user_dict)
                self.set_action({'action': arg, 'arguments': {}, 'context': the_context})
        if kwargs.get('gathering', False):
            self.next_action = None
        if first_is_plain:
            self.arguments = None

    def set_action(self, data):
        if (not hasattr(self, 'name')) or self.name is None:
            if isinstance(data, dict) and 'action' in data and (len(data) == 1 or 'arguments' in data):
                self.name = data['action']
                self.arguments = data.get('arguments', {})
                self.context = data.get('context', {})
            else:
                raise DAError("force_ask: invalid parameter " + repr(data))
        self.next_action.append(data)


class DAErrorNoEndpoint(DAError):
    pass


class DAErrorMissingVariable(DAError):

    def __init__(self, value, variable=None, code=501):
        self.value = value
        self.variable = variable
        self.error_code = code
        super().__init__(value)


class DAErrorCompileError(DAError):
    pass


class MandatoryQuestion(Exception):

    def __init__(self):
        self.value = 'Mandatory Question'
        super().__init__()

    def __str__(self):
        return str(self.value)


class QuestionError(Exception):

    def __init__(self, *pargs, **kwargs):
        if len(pargs) >= 1:
            self.question = pargs[0]
        elif 'question' in kwargs:
            self.question = kwargs['question']
        else:
            self.question = "Question not specified"
        if len(pargs) >= 2:
            self.subquestion = pargs[1]
        elif 'subquestion' in kwargs:
            self.subquestion = kwargs['subquestion']
        else:
            self.subquestion = None
        if len(pargs) >= 3:
            self.url = pargs[2]
        elif 'url' in kwargs:
            self.url = kwargs['url']
        else:
            self.url = None
        if 'show_leave' in kwargs:
            self.show_leave = kwargs['show_leave']
        else:
            self.show_leave = None
        if 'show_exit' in kwargs:
            self.show_exit = kwargs['show_exit']
        else:
            self.show_exit = None
        if 'reload' in kwargs:
            self.reload = kwargs['reload']
        else:
            self.reload = None
        if 'show_restart' in kwargs:
            self.show_restart = kwargs['show_restart']
        else:
            self.show_restart = None
        if 'buttons' in kwargs:
            self.buttons = kwargs['buttons']
        else:
            self.buttons = None
        if 'dead_end' in kwargs:
            self.dead_end = kwargs['dead_end']
        else:
            self.dead_end = None
        super().__init__()

    def __str__(self):
        return str(self.question)


class BackgroundResponseError(Exception):

    def __init__(self, *pargs, **kwargs):
        if len(pargs) > 0 and len(kwargs) > 0:
            self.backgroundresponse = {'pargs': list(pargs), 'kwargs': kwargs}
        elif len(pargs) > 1:
            self.backgroundresponse = list(pargs)
        elif len(pargs) == 1:
            self.backgroundresponse = pargs[0]
        else:
            self.backgroundresponse = kwargs
        if 'sleep' in kwargs:
            self.sleep = kwargs['sleep']
        super().__init__()

    def __str__(self):
        if hasattr(self, 'backgroundresponse'):
            return str(self.backgroundresponse)
        return "A BackgroundResponseError exception was thrown"


class BackgroundResponseActionError(Exception):

    def __init__(self, *pargs, **kwargs):
        self.action = {'arguments': {}}
        if len(pargs) == 0:
            self.action['action'] = None
        else:
            self.action['action'] = pargs[0]
        for key, val in kwargs.items():
            self.action['arguments'][key] = val
        super().__init__()

    def __str__(self):
        if hasattr(self, 'action'):
            return str(self.action)
        return "A BackgroundResponseActionError exception was thrown"


class ResponseError(Exception):

    def __init__(self, *pargs, **kwargs):
        if len(pargs) == 0 and not ('response' in kwargs or 'binaryresponse' in kwargs or 'all_variables' in kwargs or 'file' in kwargs or 'url' in kwargs or 'null' in kwargs):
            self.response = "Empty Response"
        if len(pargs) > 0:
            self.response = pargs[0]
        elif 'response' in kwargs:
            self.response = kwargs['response']
        elif 'binaryresponse' in kwargs:
            self.binaryresponse = kwargs['binaryresponse']
        elif 'file' in kwargs:
            self.filename = kwargs['file']
        elif 'url' in kwargs:
            self.url = kwargs['url']
        elif 'null' in kwargs:
            self.nullresponse = kwargs['null']
        if 'response_code' in kwargs and kwargs['response_code'] is not None:
            self.response_code = kwargs['response_code']
        if 'sleep' in kwargs:
            self.sleep = kwargs['sleep']
        if 'all_variables' in kwargs:
            self.all_variables = kwargs['all_variables']
            if 'include_internal' in kwargs:
                self.include_internal = kwargs['include_internal']
        if 'content_type' in kwargs:
            self.content_type = kwargs['content_type']
        super().__init__()

    def __str__(self):
        if hasattr(self, 'response'):
            return str(self.response)
        return "A ResponseError exception was thrown"


class CommandError(Exception):

    def __init__(self, *pargs, **kwargs):
        if len(pargs) > 0:
            self.return_type = pargs[0]
        elif 'type' in kwargs:
            self.return_type = kwargs['type']
        else:
            self.return_type = "exit"
        self.url = kwargs.get('url', '')
        self.sleep = kwargs.get('sleep', None)
        super().__init__()

    def __str__(self):
        return str(self.return_type)


class DAWebError(Exception):

    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self, key, val)
        super().__init__()



import copy
import io
import os
import multiprocessing
from concurrent import futures
from enum import Enum
from pikepdf import Pdf
from reportlab.pdfgen import canvas


class Area(Enum):
    TOP_LEFT = 0
    TOP_RIGHT = 1
    BOTTOM_RIGHT = 2
    BOTTOM_LEFT = 3


class RedactionStyle(Enum):
    SOLID = ((0, 0, 0), (0, 0, 0), (1, 1, 1))
    OUTLINE = ((0, 0, 0), (1, 1, 1), (0, 0, 0))

    def __init__(self, stroke, fill, text):
        self.stroke = stroke
        self.fill = fill
        self.text = text


class Marisol:

    def __init__(self, prefix, fill, start, area=Area.BOTTOM_RIGHT):
        """
        Marisol Base Class - A collection of documents to be bates numbered.

        Args:
            prefix (str): Bates number prefix
            fill (int): Length for zero-filling
            start (int): Starting bates number
            area (Area): Area in which to place the bates number.
        """
        self.prefix = prefix
        self.fill = fill
        self.start = start
        self.area = area

        self.index = 0
        self.number = 0

        self.documents = []
        self.overwrite = False

    def __getitem__(self, key):
        return self.documents[key]

    def __len__(self):
        return len(self.documents)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self):
            raise StopIteration
        self.index += 1
        return self.documents[self.index-1]

    def _save_document(self, document):
        """
        Internal method called by thread pool executor.

        Args:
            document (Document):  The document to save.

        Returns:
            (str, bool): The file name saved to and success or failure.
        """
        try:
            filename = document.save(overwrite=self.overwrite)
        except FileExistsError:
            return "EXISTS", False
        else:
            return filename, True

    def append(self, file):
        """
        Add a document to the collection.

        Args:
            file (str or file-like object): PDF file or file name to add.

        Returns:
            marisol.Marisol: The current Marisol instance.
        """
        d = Document(file, self.prefix, self.fill, self.start+self.number, self.area)
        self.number += len(d)
        self.documents.append(d)
        return self

    def save(self, overwrite=False, threads=multiprocessing.cpu_count()*6):
        """Save all documents using a thread pool executor

        Args:
            overwrite (bool, optional): Switch to allow overwriting of existing files.
            threads (int, optional): The number of threads to use when processing.  Defaults to the number of cores
                times six.

        Returns:
            list: each file name and true or false indicating success or failure
        """
        self.overwrite = overwrite
        with futures.ThreadPoolExecutor(threads) as executor:
            results = executor.map(self._save_document, self)
        return list(results)


class Document:

    def __init__(self, file, prefix, fill, start, area):
        """
        Represents a document to be numbered.

        Args:
            file (str): PDF file associated with this document.
            prefix (str): Bates number prefix.
            fill (int): Length to zero-pad number to.
            start (int): Number to start with.
            area (Area): Area on the document where the number should be drawn
        """
        self.reader = Pdf.open(file)
        self.prefix = prefix
        self.fill = fill
        self.start = copy.copy(start)
        self.area = area

        self.overlays = {x: None for x in Area}
        self.overlays[area] = BatesOverlay(None, self.area)

        self.index = 0

        self.pages = []
        for num, page in enumerate(self.reader.pages):
            p = Page(self, page, self.prefix, self.fill, self.start + num)
            self.pages.append(p)

    def __getitem__(self, k):
        return self.pages[k]

    def __len__(self):
        return len(self.reader.pages)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self):
            raise StopIteration
        self.index += 1
        return self.pages[self.index-1]

    def __str__(self):
        return "{begin} - {end}".format(begin=self.begin, end=self.end)

    @property
    def begin(self):
        """
        Beginning bates number for this document.

        Returns:
            str: Bates number of the first page of the document.
        """
        num = str(self.start)
        num = num.zfill(self.fill)
        return "{prefix}{num}".format(prefix=self.prefix, num=num)

    @property
    def end(self):
        """
        Ending bates number for the document

        Returns:
            str: Bates number of the last page of the document.
        """
        num = str(self.start+len(self)-1)
        num = num.zfill(self.fill)
        return "{prefix}{num}".format(prefix=self.prefix, num=num)

    def save(self, filename=None, overwrite=False):
        """
        Applies the bates numbers and saves to file.

        Args:
            filename (str): Path where the PDF should be saved.
            overwrite (bool): Switch to allow overwriting of existing files.

        Returns:
            str: Path where the file was saved.

        Raises:
            FileExistsError: When the file already exists and overwrite is not enabled.
        """
        filename = filename or "{begin}.pdf".format(begin=self.begin)

        if os.path.exists(filename) and not overwrite:
            raise FileExistsError("PDF file {} already exists and overwrite is disabled.".format(filename))

        for page in self:
            page.apply()
        self.reader.save(filename)
        return filename

    def add_overlay(self, overlay):
        """
        Add an overlay to the page in addition to the bates stamp.

        Args:
            overlay (Marisol.GenericTextOverlay):  Overlay to apply

        Raises:
            ValueError:  When area is already reserved for Bates Stamp
        """
        area = overlay.area
        if isinstance(self.overlays[area], BatesOverlay):
            raise ValueError("Area {} is already reserved for bates stamp.".format(area))
        self.overlays[area] = overlay
        return self


class Page:

    def __init__(self, document, page, prefix, fill, start):
        """
        Represents a page within a document that will be bates numbered.

        Args:
            document (Marisol.Document):  Parent document
            page (pikepdf.Page): PDF page associated with this page
            prefix (str): Bates number prefix.
            fill (int): Length to zero-pad number to.
            start (int): Number to start with.
        """
        self.document = document
        self.page = page
        self.prefix = prefix
        self.fill = fill
        self.start = start

        self.height = float(self.page.mediabox[3])
        self.width = float(self.page.mediabox[2])

        self.canvas_file = io.BytesIO()
        self.canvas = canvas.Canvas(self.canvas_file, pagesize=(self.width, self.height))

        self.redactions = []

    def __str__(self):
        return self.number

    def add_redaction(self, redaction):
        """
        Add a redaction to the page.

        Args:
            redaction (Marisol.Redaction):  Redaction to add to the page.
        """
        position = redaction.position
        size = redaction.size
        if position[0]+size[0] > self.width or position[1]+size[1] > self.height:
            raise OutsideBoundariesError("Redaction with position {} and \
            size {} is outside of page ({},{})".format(position, size, self.width, self.height))
        self.redactions.append(redaction)
        return self

    def apply(self):
        """
        Applies all requested overlays to the page

        Returns:
            bool
        """
        for overlay in self.document.overlays.values():
            if isinstance(overlay, BatesOverlay):
                overlay.text = self.number
                overlay.apply(self.canvas)
            elif isinstance(overlay, GenericTextOverlay):
                overlay.apply(self.canvas)

        for redaction in self.redactions:
            redaction.apply(self.canvas)

        self.canvas.showPage()
        self.canvas.save()

        self.canvas_file.seek(0)
        reader = Pdf.open(self.canvas_file)
        overlay_page = reader.pages[0]
        self.page.add_overlay(overlay_page)
        return True

    @property
    def number(self):
        """
        The bates number for the page.

        Returns:
            str: Bates number.
        """
        num = str(self.start)
        num = num.zfill(self.fill)
        return "{prefix}{num}".format(prefix=self.prefix, num=num)


class GenericTextOverlay:

    def __init__(self, text, area):
        self.text = text
        self.area = area

    def apply(self, c):
        """
        Applies the bates number to a canvas.

        Args:
             c (canvas.Canvas): canvas to apply the overlay to
        """
        position_left, position_bottom = self.position(c)
        c.drawString(position_left, position_bottom, self.text)

    def position(self, c):
        """
        Get the appropriate position on the page for the current text given an area.

        Args:
            c (canvas.Canvas): Page to get the positioning for

        Returns:
            tuple: the position
        """
        if self.area in [Area.TOP_LEFT, Area.TOP_RIGHT]:  # top
            from_bottom = c._pagesize[1]-15  # 15 down from height of page
        elif self.area in [Area.BOTTOM_LEFT, Area.BOTTOM_RIGHT]:  # bottom
            from_bottom = 15  # 15 up from bottom of page

        if self.area in [Area.TOP_LEFT, Area.BOTTOM_LEFT]:  # left
            from_left = 15
        elif self.area in [Area.TOP_RIGHT, Area.BOTTOM_RIGHT]:  # right
            offset = 15  # initial offset
            offset += c.stringWidth(self.text)  # offset for text length
            from_left = c._pagesize[0]-offset

        return from_left, from_bottom


class BatesOverlay(GenericTextOverlay):
    pass


class StaticOverlay(GenericTextOverlay):
    pass


class Redaction:

    def __init__(self, position, size, text=None, style=RedactionStyle.SOLID):
        """

        Args:
            position (tuple): from-left and from-bottom position to draw redaction at in points
            size (tuple): width and height of the redaction in points.
            text (str):
            style (Marisol.RedactionStyle):
        """
        self.position = position
        self.size = size

        # center of the drawn redaction (from-left, from-bottom)
        self.center = (self.position[0]+self.size[0]/2,
                       (self.position[1]+self.size[1]/2)-5.0)

        self.text = text
        self.style = style

    def apply(self, c):
        """
        Apply the redaction to a canvas.

        Args:
            c (Canvas): canvas to apply the redaction to.
        """

        c.setFont("Helvetica", 10)

        c.setStrokeColorRGB(*self.style.stroke)
        c.setFillColorRGB(*self.style.fill)

        c.rect(*self.position+self.size, fill=1)

        if self.text is not None:
            c.setStrokeColorRGB(*self.style.text)
            c.setFillColorRGB(*self.style.text)
            c.drawCentredString(self.center[0], self.center[1], self.text)


class OutsideBoundariesError(ValueError):
    """Raised when an item is drawn outside the page boundaries."""



import sys
import os
import re
separator = re.compile(r' *[,;] *')

if __name__ == "__main__":
    import docassemble.base.config
    docassemble.base.config.load(arguments=sys.argv)
    from docassemble.base.config import daconfig, parse_redis_uri
    if 'timezone' in daconfig and daconfig['timezone'] is not None:
        print('export TIMEZONE="' + str(daconfig['timezone']) + '"')
    if 'os locale' in daconfig and daconfig['os locale'] is not None:
        print('export LOCALE="' + str(daconfig['os locale']) + '"')
    else:
        print('export LOCALE="en_US.UTF-8 UTF-8"')
    if 'web server' in daconfig and isinstance(daconfig['web server'], str):
        print('export DAWEBSERVER="' + daconfig['web server'] + '"')
    else:
        print('export DAWEBSERVER="nginx"')
    if '--limited' in sys.argv:
        sys.exit(0)
    if 'other os locales' in daconfig and isinstance(daconfig['other os locales'], list):
        print('declare -a OTHERLOCALES')
        print('export OTHERLOCALES')
        indexno = 0
        for locale in daconfig['other os locales']:
            print('OTHERLOCALES[' + str(indexno) + ']=' + repr(str(locale)))
            indexno += 1
    else:
        other_locales_variable = os.getenv('OTHERLOCALES', None)
        if other_locales_variable is not None and other_locales_variable != 'null':
            print('declare -a OTHERLOCALES')
            print('export OTHERLOCALES')
            indexno = 0
            for locale in map(lambda x: x.strip(), separator.split(other_locales_variable)):
                print('OTHERLOCALES[' + str(indexno) + ']=' + repr(str(locale)))
                indexno += 1
    max_content_length = daconfig.get('maximum content length', 16 * 1024 * 1024)
    if isinstance(max_content_length, (int, type(None))):
        if max_content_length is None or max_content_length <= 0:
            print('export DAMAXCONTENTLENGTH=0')
        else:
            print('export DAMAXCONTENTLENGTH=' + str(max_content_length))
    else:
        print('DAMAXCONTENTLENGTH=' + str(16 * 1024 * 1024))
    if 'celery processes' in daconfig and isinstance(daconfig['celery processes'], int):
        print('DACELERYWORKERS=' + str(daconfig['celery processes']))
    if 'debian packages' in daconfig and isinstance(daconfig['debian packages'], list) and daconfig.get('ubuntu packages', None) is not None:
        daconfig['ubuntu packages'] = daconfig['debian packages']
        del daconfig['debian packages']
    if 'debian packages' in daconfig and isinstance(daconfig['debian packages'], list):
        print('declare -a PACKAGES')
        print('export PACKAGES')
        indexno = 0
        for package in daconfig['debian packages']:
            print('PACKAGES[' + str(indexno) + ']=' + repr(str(package)))
            indexno += 1
    else:
        packages_variable = os.getenv('PACKAGES', None)
        if packages_variable is not None and packages_variable != 'null':
            print('declare -a PACKAGES')
            print('export PACKAGES')
            indexno = 0
            for package in map(lambda x: x.strip(), separator.split(packages_variable)):
                print('PACKAGES[' + str(indexno) + ']=' + repr(str(package)))
                indexno += 1
    if 'python packages' in daconfig and isinstance(daconfig['python packages'], list):
        print('declare -a PYTHONPACKAGES')
        print('export PYTHONPACKAGES')
        indexno = 0
        for package in daconfig['python packages']:
            print('PYTHONPACKAGES[' + str(indexno) + ']=' + repr(str(package)))
            indexno += 1
    else:
        packages_variable = os.getenv('PYTHONPACKAGES', None)
        if packages_variable is not None and packages_variable != 'null':
            print('declare -a PYTHONPACKAGES')
            print('export PYTHONPACKAGES')
            indexno = 0
            for package in map(lambda x: x.strip(), separator.split(packages_variable)):
                print('PYTHONPACKAGES[' + str(indexno) + ']=' + repr(str(package)))
                indexno += 1
    if 'db' in daconfig:
        if 'prefix' in daconfig['db'] and daconfig['db']['prefix'] is not None:
            if daconfig['db']['prefix'].startswith('postgresql'):
                print('export DBTYPE="postgresql"')
            elif daconfig['db']['prefix'].startswith('mysql'):
                print('export DBTYPE="mysql"')
            else:
                print('export DBTYPE="other"')
            print('export DBPREFIX="' + str(daconfig['db']['prefix']) + '"')
        if 'name' in daconfig['db'] and daconfig['db']['name'] is not None:
            print('export DBNAME="' + str(daconfig['db']['name']) + '"')
        if 'user' in daconfig['db'] and daconfig['db']['user'] is not None:
            print('export DBUSER="' + str(daconfig['db']['user']) + '"')
        if 'password' in daconfig['db'] and daconfig['db']['password'] is not None:
            print('export DBPASSWORD="' + str(daconfig['db']['password']) + '"')
        if 'host' in daconfig['db'] and daconfig['db']['host'] is not None:
            print('export DBHOST="' + str(daconfig['db']['host']) + '"')
        if 'port' in daconfig['db'] and daconfig['db']['port'] is not None:
            print('export DBPORT="' + str(daconfig['db']['port']) + '"')
        if 'table prefix' in daconfig['db'] and daconfig['db']['table prefix'] is not None:
            print('export DBTABLEPREFIX="' + str(daconfig['db']['table prefix']) + '"')
        if 'backup' in daconfig['db'] and daconfig['db']['backup'] is not None:
            print('export DBBACKUP="' + ('true' if daconfig['db']['backup'] else 'false') + '"')
        if 'ssl mode' in daconfig['db'] and daconfig['db']['ssl mode'] is not None:
            print('export DBSSLMODE="' + str(daconfig['db']['ssl mode']) + '"')
        if 'ssl cert' in daconfig['db'] and daconfig['db']['ssl cert'] is not None:
            print('export DBSSLCERT="' + str(daconfig['db']['ssl cert']) + '"')
        if 'ssl key' in daconfig['db'] and daconfig['db']['ssl key'] is not None:
            print('export DBSSLKEY="' + str(daconfig['db']['ssl key']) + '"')
        if 'ssl root cert' in daconfig['db'] and daconfig['db']['ssl root cert'] is not None:
            print('export DBSSLROOTCERT="' + str(daconfig['db']['ssl root cert']) + '"')
    if daconfig['supervisor'].get('username', None):
        print('export DASUPERVISORUSERNAME=' + str(daconfig['supervisor']['username']))
        print('export DASUPERVISORPASSWORD=' + str(daconfig['supervisor']['password']))
        print('export DASUPERVISOROPTS="--username ' + str(daconfig['supervisor']['username']) + ' --password ' + str(daconfig['supervisor']['password']) + ' "')
    else:
        print('export DASUPERVISORUSERNAME=""')
        print('export DASUPERVISORPASSWORD=""')
        print('export DASUPERVISOROPTS=""')
    if 'update on start' in daconfig:
        if daconfig['update on start'] is False:
            print('export DAUPDATEONSTART=false')
        elif daconfig['update on start'] == 'initial':
            print('export DAUPDATEONSTART=initial')
    if 'allow updates' in daconfig and daconfig['allow updates'] is False:
        print('export DAALLOWUPDATES=false')
    if 'allow configuration editing' in daconfig and daconfig['allow configuration editing'] is False:
        print('export DAALLOWCONFIGURATIONEDITING=false')
    if 'enable playground' in daconfig and daconfig['enable playground'] is False:
        print('export DAENABLEPLAYGROUND=false')
    if 'allow log viewing' in daconfig and daconfig['allow log viewing'] is False:
        print('export DAALLOWLOGVIEWING=false')
    if 'root owned' in daconfig and daconfig['root owned'] is True:
        print('export DAROOTOWNED=true')
    if 'read only file system' in daconfig and daconfig['read only file system'] is True:
        print('export DAREADONLYFILESYSTEM=true')
    if 'expose websockets' in daconfig and daconfig['expose websockets']:
        print('export DAEXPOSEWEBSOCKETS=true')
    if 'websockets ip' in daconfig and daconfig['websockets ip']:
        print('export DAWEBSOCKETSIP="' + str(daconfig['websockets ip']) + '"')
    if 'http port' in daconfig and daconfig['http port']:
        print('export PORT="' + str(daconfig['http port']) + '"')
    if 'stable version' in daconfig and daconfig['stable version']:
        print('export DASTABLEVERSION=true')
    if 'nginx ssl protocols' in daconfig and daconfig['nginx ssl protocols']:
        print('export DASSLPROTOCOLS=' + str(daconfig['nginx ssl protocols']))
    if 'websockets port' in daconfig and daconfig['websockets port']:
        print('export DAWEBSOCKETSPORT=' + str(daconfig['websockets port']))
    else:
        print('export DAWEBSOCKETSPORT=5000')
    if 'redis' in daconfig and daconfig['redis'] is not None:
        print('export REDIS="' + str(daconfig['redis']) + '"')
        (redis_host, redis_port, redis_username, redis_password, redis_offset, redis_cli, ssl_opts) = parse_redis_uri()
        print('export REDISCLI="' + str(redis_cli) + '"')
    if 'rabbitmq' in daconfig and daconfig['rabbitmq'] is not None:
        print('export RABBITMQ="' + str(daconfig['rabbitmq']) + '"')
    if 'backup days' in daconfig:
        try:
            days = int(daconfig['backup days'])
            assert days >= 0
        except:
            days = 14
        print('export DABACKUPDAYS="' + str(days) + '"')
    else:
        print('export DABACKUPDAYS="14"')
    if 'backup file storage' in daconfig and not daconfig['backup file storage']:
        print('export BACKUPFILESTORAGE=false')
    else:
        print('export BACKUPFILESTORAGE=true')
    if 'enable unoconv' in daconfig and daconfig['enable unoconv'] is True:
        print('export ENABLEUNOCONV=true')
    else:
        print('export ENABLEUNOCONV=false')
    if 's3' in daconfig:
        s4_options = []
        if ('enable' in daconfig['s3'] and daconfig['s3']['enable']) or ('enable' not in daconfig['s3'] and 'bucket' in daconfig['s3'] and daconfig['s3']['bucket'] is not None):
            print('export S3ENABLE=true')
        else:
            print('export S3ENABLE=false')
        if 'access key id' in daconfig['s3'] and daconfig['s3']['access key id'] is not None:
            print('export S3ACCESSKEY="' + str(daconfig['s3']['access key id']) + '"')
            print('export AWS_ACCESS_KEY_ID="' + str(daconfig['s3']['access key id']) + '"')
            print('export S3_ACCESS_KEY="' + str(daconfig['s3']['access key id']) + '"')
        if 'secret access key' in daconfig['s3'] and daconfig['s3']['secret access key'] is not None:
            print('export S3SECRETACCESSKEY="' + str(daconfig['s3']['secret access key']) + '"')
            print('export AWS_SECRET_ACCESS_KEY="' + str(daconfig['s3']['secret access key']) + '"')
            print('export S3_SECRET_KEY="' + str(daconfig['s3']['secret access key']) + '"')
        if 'bucket' in daconfig['s3'] and daconfig['s3']['bucket'] is not None:
            print('export S3BUCKET="' + str(daconfig['s3']['bucket']) + '"')
        if 'region' in daconfig['s3'] and daconfig['s3']['region'] is not None:
            print('export S3REGION="' + str(daconfig['s3']['region']) + '"')
            print('export AWS_DEFAULT_REGION="' + str(daconfig['s3']['region']) + '"')
        if 'endpoint url' in daconfig['s3'] and daconfig['s3']['endpoint url'] is not None:
            print('export S3ENDPOINTURL="' + str(daconfig['s3']['endpoint url']) + '"')
            s4_options.append('--endpoint-url=\\"' + str(daconfig['s3']['endpoint url']) + '\\"')
        if 'server side encryption' in daconfig['s3'] and isinstance(daconfig['s3']['server side encryption'], dict):
            if 'algorithm' in daconfig['s3']['server side encryption'] and daconfig['s3']['server side encryption']['algorithm'] is not None:
                print('export S3_SSE_ALGORITHM="' + str(daconfig['s3']['server side encryption']['algorithm']).strip() + '"')
                s4_options.append('--API-ServerSideEncryption=\\"' + str(daconfig['s3']['server side encryption']['algorithm']).strip() + '\\"')
            if 'customer algorithm' in daconfig['s3']['server side encryption'] and daconfig['s3']['server side encryption']['customer algorithm'] is not None:
                print('export S3_SSE_CUSTOMER_ALGORITHM="' + str(daconfig['s3']['server side encryption']['customer algorithm']).strip() + '"')
                s4_options.append('--API-SSECustomerAlgorithm=\\"' + str(daconfig['s3']['server side encryption']['customer algorithm']).strip() + '\\"')
            if 'customer key' in daconfig['s3']['server side encryption'] and daconfig['s3']['server side encryption']['customer key'] is not None:
                print('export S3_SSE_CUSTOMER_KEY="' + str(daconfig['s3']['server side encryption']['customer key']).strip() + '"')
                s4_options.append('--API-SSECustomerKey=\\"' + str(daconfig['s3']['server side encryption']['customer key']).strip() + '\\"')
            if 'KMS key ID' in daconfig['s3']['server side encryption'] and daconfig['s3']['server side encryption']['KMS key ID'] is not None:
                print('export S3_SSE_KMS_KEY_ID="' + str(daconfig['s3']['server side encryption']['KMS key ID']).strip() + '"')
                s4_options.append('--API-SSEKMSKeyId=\\"' + str(daconfig['s3']['server side encryption']['KMS key ID']).strip() + '\\"')
        if len(s4_options) > 0:
            print('export S4CMD_OPTS="' + " ".join(s4_options) + '"')
    if 'azure' in daconfig:
        if ('enable' in daconfig['azure'] and daconfig['azure']['enable']) or ('enable' not in daconfig['azure'] and 'container' in daconfig['azure'] and daconfig['azure']['container'] is not None and 'account name' in daconfig['azure'] and daconfig['azure']['account name'] is not None):
            print('export AZUREENABLE=true')
            print('export AZURE_STORAGE_KEY="' + str(daconfig['azure'].get('account key', '')) + '"')
            print('export AZURE_STORAGE_ACCOUNT="' + str(daconfig['azure']['account name']) + '"')
            print('export AZURE_STORAGE_AUTH_MODE=key')
        else:
            print('export AZUREENABLE=false')
        if 'connection string' in daconfig['azure'] and daconfig['azure']['connection string'] is not None:
            print('export AZURECONNECTIONSTRING="' + str(daconfig['azure']['connection string']) + '"')
        if 'account name' in daconfig['azure'] and daconfig['azure']['account name'] is not None:
            print('export AZUREACCOUNTNAME="' + str(daconfig['azure']['account name']) + '"')
        if 'account key' in daconfig['azure'] and daconfig['azure']['account key'] is not None:
            print('export AZUREACCOUNTKEY="' + str(daconfig['azure']['account key']) + '"')
        if 'container' in daconfig['azure'] and daconfig['azure']['container'] is not None:
            print('export AZURECONTAINER="' + str(daconfig['azure']['container']) + '"')
    if 'ec2' in daconfig and daconfig['ec2']:
        print('export EC2=true')
    if 'collect statistics' in daconfig and daconfig['collect statistics']:
        print('export COLLECTSTATISTICS=true')
    if 'kubernetes' in daconfig and daconfig['kubernetes']:
        print('export KUBERNETES=true')
    if 'log server' in daconfig and daconfig['log server'] is not None:
        print('export LOGSERVER="' + str(daconfig['log server']) + '"')
    if 'log' in daconfig and daconfig['log'] is not None:
        print('export LOGDIRECTORY="' + str(daconfig['log']) + '"')
    if 'use https' in daconfig and daconfig['use https']:
        print('export USEHTTPS=true')
    else:
        print('export USEHTTPS=false')
    if 'use cloud urls' in daconfig and daconfig['use cloud urls']:
        print('export USECLOUDURLS=true')
    else:
        print('export USECLOUDURLS=false')
    if 'use minio' in daconfig and daconfig['use minio']:
        print('export USEMINIO=true')
    else:
        print('export USEMINIO=false')
    if 'use lets encrypt' in daconfig and daconfig['use lets encrypt']:
        print('export USELETSENCRYPT=true')
    else:
        print('export USELETSENCRYPT=false')
    if 'behind https load balancer' in daconfig and daconfig['behind https load balancer']:
        print('export BEHINDHTTPSLOADBALANCER=true')
    else:
        print('export BEHINDHTTPSLOADBALANCER=false')
    if 'lets encrypt email' in daconfig and daconfig['lets encrypt email'] is not None:
        print('export LETSENCRYPTEMAIL="' + str(daconfig['lets encrypt email']) + '"')
    if 'external hostname' in daconfig and daconfig['external hostname'] is not None:
        print('export DAHOSTNAME="' + str(daconfig['external hostname']) + '"')
    if 'root' in daconfig and daconfig['root'] is not None:
        print('export POSTURLROOT="' + str(daconfig['root']) + '"')
        print('export WSGIROOT="' + str(re.sub(r'^(.+)/$', r'\1', daconfig['root'])) + '"')
    else:
        print('export POSTURLROOT="/"')
        print('export WSGIROOT="/"')
    if 'server administrator email' in daconfig and daconfig['server administrator email']:
        print('export SERVERADMIN="' + str(daconfig['server administrator email']) + '"')
    else:
        print('export SERVERADMIN="webmaster@localhost"')
    if 'web server timeout' in daconfig and daconfig['web server timeout'] is not None:
        print('export DATIMEOUT="' + str(daconfig['web server timeout']) + '"')
    if 'pip index url' in daconfig and daconfig['pip index url'] is not None and daconfig['pip index url'] != '':
        print('export PIPINDEXURL="' + str(daconfig['pip index url']) + '"')
    if 'pip extra index urls' in daconfig and daconfig['pip extra index urls'] is not None and daconfig['pip extra index urls'] != '':
        print('export PIPEXTRAINDEXURLS="' + str(daconfig['pip extra index urls']) + '"')
    sys.exit(0)
